@@ -60,11 +60,13 @@ module Network.CGI (
   , outputNotFound, outputMethodNotAllowed, outputInternalServerError
   -- * Input
   , getInput, getInputFPS, readInput
-  , getInputs, getInputNames
-  , getMultiInput
+  , getBody, getBodyFPS
+  , getInputs, getInputsFPS, getInputNames
+  , getMultiInput, getMultiInputFPS
   , getInputFilename, getInputContentType
   -- * Environment
   , getVar, getVarWithDefault, getVars
+  -- * Request information
   , serverName, serverPort
   , requestMethod, pathInfo
   , pathTranslated, scriptName
@@ -75,6 +77,11 @@ module Network.CGI (
   , requestHeader
   -- * Program and request URI
   , progURI, queryURI, requestURI
+  -- * Content negotiation
+  , Acceptable, Accept
+  , Charset(..), ContentEncoding(..), Language(..)
+  , requestAccept, requestAcceptCharset, requestAcceptEncoding, requestAcceptLanguage
+  , negotiate
   -- * Content type
   , ContentType(..), showContentType, parseContentType
   -- * Cookies
@@ -83,6 +90,8 @@ module Network.CGI (
   , setCookie, deleteCookie
   -- * URL encoding
   , formEncode, urlEncode, formDecode, urlDecode
+  -- * Compatibility with the old API
+  , module Network.CGI.Compat
   ) where
 
 import Control.Exception (Exception(..))
@@ -102,10 +111,11 @@ import Data.ByteString.Lazy.Char8 (ByteString)
 
 import Network.CGI.Cookie (Cookie(..), showCookie, newCookie, findCookie)
 import qualified Network.CGI.Cookie as Cookie (deleteCookie)
-import Network.CGI.RFC822Headers (ContentType(..), 
-                                  parseContentType, showContentType)
+import Network.CGI.Accept
+import Network.CGI.Header
 import Network.CGI.Monad
 import Network.CGI.Protocol
+import Network.CGI.Compat
 
 import Text.XHtml (Html, renderHtml, header, (<<), thetitle, (+++), 
                    body, h1, paragraph, hr, address)
@@ -193,9 +203,17 @@ outputError :: (MonadCGI m, MonadIO m) =>
 outputError c m es = 
       do logCGI $ show (c,m,es)
          setStatus c m
-         setHeader "Content-type" "text/html; charset=ISO-8859-1"
-         page <- errorPage c m es 
-         output $ renderHtml page
+         let textType = ContentType "text" "plain" [("charset","ISO-8859-1")]
+             htmlType = ContentType "text" "html"  [("charset","ISO-8859-1")]
+         cts <- liftM (negotiate [htmlType,textType]) requestAccept
+         case cts of
+           ct:_ | ct == textType -> 
+                do setHeader "Content-type" (showContentType textType)
+                   text <- errorText c m es
+                   output text
+           _ -> do setHeader "Content-type" (showContentType htmlType)
+                   page <- errorPage c m es 
+                   output $ renderHtml page
 
 -- | Create an HTML error page.
 errorPage :: MonadCGI m => 
@@ -215,6 +233,13 @@ errorPage c m es =
                   +++ body << (h1 << tit +++ map (paragraph <<) es 
                                +++ hr +++ address << sig)
 
+errorText :: MonadCGI m => 
+             Int      -- ^ Status code
+          -> String   -- ^ Status message
+          -> [String] -- ^ Error information
+          -> m String
+errorText c m es = return $ unlines $ (show c ++ " " ++ m) : es
+
 --
 -- * Specific HTTP errors
 --
@@ -265,6 +290,9 @@ getVars :: MonadCGI m =>
            m [(String,String)]
 getVars = liftM Map.toList $ cgiGet cgiVars
 
+--
+-- * Request information
+--
 
 -- | The server\'s hostname, DNS alias, or IP address as it would 
 --   appear in self-referencing URLs.
@@ -362,6 +390,24 @@ requestHeader :: MonadCGI m => String -> m (Maybe String)
 requestHeader name = getVar var
   where var = "HTTP_" ++ map toUpper (replace '-' '_' name)
 
+--
+-- * Content negotiation
+--
+
+requestHeaderValue :: (MonadCGI m, HeaderValue a) => String -> m (Maybe a)
+requestHeaderValue h = liftM (>>= parseM parseHeaderValue h) $ requestHeader h
+
+requestAccept :: MonadCGI m => m (Maybe (Accept ContentType))
+requestAccept = requestHeaderValue "Accept"
+
+requestAcceptCharset :: MonadCGI m => m (Maybe (Accept Charset))
+requestAcceptCharset = requestHeaderValue "Accept-Charset"
+
+requestAcceptEncoding :: MonadCGI m => m (Maybe (Accept ContentEncoding))
+requestAcceptEncoding = requestHeaderValue "Accept-Encoding"
+
+requestAcceptLanguage :: MonadCGI m => m (Maybe (Accept Language))
+requestAcceptLanguage = requestHeaderValue "Accept-Language"
 
 --
 -- * Program and request URI
@@ -479,8 +525,15 @@ getMultiInput :: MonadCGI m =>
                  String -- ^ The name of the variable.
               -> m [String] -- ^ The values of the variable,
                             -- or the empty list if the variable was not set.
-getMultiInput n = do is <- cgiGet cgiInputs
-                     return [BS.unpack (inputValue v) | (p,v) <- is, p == n]
+getMultiInput = liftM (map BS.unpack) . getMultiInputFPS
+
+-- | Same as 'getMultiInput' but using 'ByteString's.
+getMultiInputFPS :: MonadCGI m => 
+                    String -- ^ The name of the variable.
+                 -> m [ByteString] -- ^ The values of the variable,
+                            -- or the empty list if the variable was not set.
+getMultiInputFPS n = do is <- cgiGet cgiInputs
+                        return [inputValue v | (p,v) <- is, p == n]
 
 -- | Get the file name of an input.
 getInputFilename :: MonadCGI m =>
@@ -514,6 +567,13 @@ getInputs :: MonadCGI m => m [(String,String)]
 getInputs = do is <- cgiGet cgiInputs
                return [ (n, BS.unpack (inputValue i)) | (n,i) <- is ]
 
+-- | Get the names and values of all inputs.
+--   Note: the same name may occur more than once in the output,
+--   if there are several values for the name.
+getInputsFPS :: MonadCGI m => m [(String,ByteString)]
+getInputsFPS = do is <- cgiGet cgiInputs
+                  return [ (n, inputValue i) | (n,i) <- is ]
+
 -- | Get the names of all input variables.
 getInputNames :: MonadCGI m => m [String]
 getInputNames = (sortNub . map fst) `liftM` cgiGet cgiInputs
@@ -524,6 +584,13 @@ getInputNames = (sortNub . map fst) `liftM` cgiGet cgiInputs
 getInput_ ::  MonadCGI m => String -> m (Maybe Input)
 getInput_ n = lookup n `liftM` cgiGet cgiInputs
 
+-- | Get the uninterpreted request body as a String
+getBody :: MonadCGI m => m String
+getBody = BS.unpack `liftM` cgiGet cgiRequestBody
+
+-- | Get the uninterpreted request body as lazy ByteString
+getBodyFPS :: MonadCGI m => m ByteString
+getBodyFPS = cgiGet cgiRequestBody
 
 --
 -- * Cookies
